"""
Conversion utilities for creating pyat elements
"""
import collections
import os
import re
import numpy
from warnings import warn
from typing import Optional
import sysconfig
from at import integrators
from at.lattice import AtWarning
from at.lattice import CLASS_MAP, elements as elt
from at.lattice import Particle, Element
# imports necessary in' globals()' for 'eval'
# noinspection PyUnresolvedReferences
from numpy import array, uint8  # For global namespace

_ext_suffix = sysconfig.get_config_var('EXT_SUFFIX')


def _particle(value):
    if isinstance(value, Particle):
        # Create from python: save_mat
        return value
    else:
        # Create from Matlab: load_mat
        name = value.pop('name')
        return Particle(name, **value)


class RingParam(elt.Element):
    """Private class for Matlab RingParam element

    :meta private:
    """
    # noinspection PyProtectedMember
    _BUILD_ATTRIBUTES = elt.Element._BUILD_ATTRIBUTES + ['Energy',
                                                         'Periodicity']
    _conversions = dict(elt.Element._conversions, Energy=float,
                        Periodicity=int, Particle=_particle)

    def __init__(self, family_name, energy, periodicity=1, **kwargs):
        kwargs.setdefault('Energy', energy)
        kwargs.setdefault('Periodicity', periodicity)
        kwargs.setdefault('PassMethod', 'IdentityPass')
        super(RingParam, self).__init__(family_name, **kwargs)


_alias_map = {'rbend': elt.Dipole,
              'sbend': elt.Dipole,
              'quad': elt.Quadrupole,
              'sext': elt.Sextupole,
              'rf': elt.RFCavity,
              'bpm': elt.Monitor,
              'ap': elt.Aperture,
              'ringparam': RingParam,
              'wig': elt.Wiggler}


# Matlab to Python class translation
_CLASS_MAP = dict((k.lower(), v) for k, v in CLASS_MAP.items())
_CLASS_MAP.update(_alias_map)

_PASS_MAP = {'BendLinearPass': elt.Dipole,
             'BndMPoleSymplectic4RadPass': elt.Dipole,
             'BndMPoleSymplectic4Pass': elt.Dipole,
             'QuadLinearPass': elt.Quadrupole,
             'CorrectorPass': elt.Corrector,
             'CavityPass': elt.RFCavity, 'RFCavityPass': elt.RFCavity,
             'ThinMPolePass': elt.ThinMultipole,
             'Matrix66Pass': elt.M66,
             'AperturePass': elt.Aperture,
             'GWigSymplecticPass': elt.Wiggler}

# Matlab to Python attribute translation
_param_to_lattice = {'Energy': 'energy', 'Periodicity': 'periodicity',
                     'FamName': 'name'}

# Python to Matlab class translation
_matclass_map = {'Dipole': 'Bend'}

# Python to Matlab type translation
_mattype_map = {int: float,
                numpy.ndarray: lambda attr: numpy.asanyarray(attr),
                Particle: lambda attr: attr.to_dict()}

_class_to_matfunc = {
    elt.Dipole: 'atsbend',
    elt.Bend: 'atsbend',
    elt.M66: 'atM66'}


def hasattrs(kwargs: dict, *attributes) -> bool:
    """Checks the presence of keys in a :py:class:`dict`

    Returns :py:obj:`True` if any of the ``attributes`` is in ``kwargs``

    Args:
        kwargs:     The dictionary of keyword arguments passed to the
          Element constructor.
        attributes: A list of strings, the attribute names to be checked.

    Returns:
        found (bool):   :py:obj:`True` if the element has any of the specified
          attributes.
    """
    for attribute in attributes:
        if attribute in kwargs:
            return True
    return False


def find_class(elem_dict: dict, quiet: bool = False) -> type(Element):
    """Identify the class of an element from its attributes

    Args:
        elem_dict:      The dictionary of keyword arguments passed to the
                        Element constructor.
        quiet:          Suppress the warning for non-standard classes

    Returns:
        element_class:  The guessed Class name
    """

    def low_order(key):
        polynom = numpy.array(elem_dict[key], dtype=numpy.float64).reshape(-1)
        try:
            low = numpy.where(polynom != 0.0)[0][0]
        except IndexError:
            low = -1
        return low

    class_name = elem_dict.pop('Class', '')
    try:
        return _CLASS_MAP[class_name.lower()]
    except KeyError:
        if not quiet and class_name:
            warn(AtWarning("Class '{0}' does not exist.\n"
                           "{1}".format(class_name, elem_dict)))
        fam_name = elem_dict.get('FamName', '')
        try:
            return _CLASS_MAP[fam_name.lower()]
        except KeyError:
            pass_method = elem_dict.get('PassMethod', '')
            if not quiet and not pass_method:
                warn(AtWarning("No PassMethod provided."
                               "\n{0}".format(elem_dict)))
            elif not quiet and not pass_method.endswith('Pass'):
                warn(AtWarning("Invalid PassMethod ({0}), provided pass "
                               "methods should end in 'Pass'."
                               "\n{1}".format(pass_method, elem_dict)))
            class_from_pass = _PASS_MAP.get(pass_method)
            if class_from_pass is not None:
                return class_from_pass
            else:
                length = float(elem_dict.get('Length', 0.0))
                if hasattrs(elem_dict, 'FullGap', 'FringeInt1', 'FringeInt2',
                            'gK', 'EntranceAngle', 'ExitAngle'):
                    return elt.Dipole
                elif hasattrs(elem_dict, 'Voltage', 'Frequency', 'HarmNumber',
                              'PhaseLag', 'TimeLag'):
                    return elt.RFCavity
                elif hasattrs(elem_dict, 'Periodicity'):
                    # noinspection PyProtectedMember
                    return RingParam
                elif hasattrs(elem_dict, 'Limits'):
                    return elt.Aperture
                elif hasattrs(elem_dict, 'M66'):
                    return elt.M66
                elif hasattrs(elem_dict, 'K'):
                    return elt.Quadrupole
                elif hasattrs(elem_dict, 'PolynomB', 'PolynomA'):
                    loworder = low_order('PolynomB')
                    if loworder == 1:
                        return elt.Quadrupole
                    elif loworder == 2:
                        return elt.Sextupole
                    elif loworder == 3:
                        return elt.Octupole
                    elif (pass_method.startswith('StrMPoleSymplectic4') or
                          (length > 0)):
                        return elt.Multipole
                    else:
                        return elt.ThinMultipole
                elif hasattrs(elem_dict, 'KickAngle'):
                    return elt.Corrector
                elif length > 0.0:
                    return elt.Drift
                elif hasattrs(elem_dict, 'GCR'):
                    return elt.Monitor
                elif pass_method == 'IdentityPass':
                    return elt.Marker
                else:
                    return elt.Element


def element_from_dict(elem_dict: dict, index: Optional[int] = None,
                      check: bool = True, quiet: bool = False) -> Element:
    """Builds an :py:class:`.Element` from a dictionary of attributes

    Parameters:
        elem_dict:      Dictionary of element attributes
        index:          Element index
        check:          Check the compatibility of class and PassMethod
        quiet:          Suppress the warning for non-standard classes

    Returns:
        elem (Element): new :py:class:`.Element`
    """

    # noinspection PyShadowingNames
    def sanitise_class(index, cls, elem_dict):
        """Checks that the Class and PassMethod of the element are a valid
            combination. Some Classes and PassMethods are incompatible and
            would raise errors during calculation if left, so we raise an error
            here with a more helpful message.

        Args:
            index:          element index
            cls:            Proposed class
            elem_dict:      The dictionary of keyword arguments passed to the
                            Element constructor.

        Raises:
            AttributeError: if the PassMethod and Class are incompatible.
        """
        def err(message, *args):
            location = ': ' if index is None else ' {0}: '.format(index)
            msg = ''.join(('Error in element', location,
                           'PassMethod {0} '.format(pass_method),
                           message.format(*args), '\n{0}'.format(elem_dict)))
            return AttributeError(msg)

        class_name = cls.__name__
        pass_method = elem_dict.get('PassMethod')
        if pass_method is not None:
            pass_to_class = _PASS_MAP.get(pass_method)
            length = float(elem_dict.get('Length', 0.0))
            file_name = pass_method + _ext_suffix
            file_path = os.path.join(integrators.__path__[0], file_name)
            if not os.path.isfile(os.path.realpath(file_path)):
                raise err("does not have a {0} file.".format(file_name))
            elif (pass_method == 'IdentityPass') and (length != 0.0):
                raise err("is not compatible with length {0}.", length)
            elif pass_to_class is not None:
                if not issubclass(cls, pass_to_class):
                    raise err("is not compatible with Class {0}.", class_name)
            elif issubclass(cls, (elt.Marker, elt.Monitor, RingParam)):
                if pass_method != 'IdentityPass':
                    raise err("is not compatible with Class {0}.", class_name)
            elif cls == elt.Drift:
                if pass_method != 'DriftPass':
                    raise err("is not compatible with Class {0}.", class_name)

    cls = find_class(elem_dict, quiet=quiet)
    if check:
        sanitise_class(index, cls, elem_dict)
    # Remove mandatory attributes from the keyword arguments.
    # Create list rather than generator to ensure that elements are removed
    # from elem_dict.
    elem_args = [elem_dict.pop(attr, None) for attr in cls._BUILD_ATTRIBUTES]
    element = cls(*(arg for arg in elem_args if arg is not None), **elem_dict)
    return element


def element_from_string(elem_string: str) -> Element:
    """Builds an :py:class:`.Element` from its python :py:func:`repr` string

    Parameters:
        elem_string:    String representation of an :py:class:`.Element`

    Returns:
        elem (Element): new :py:class:`.Element`
    """
    return eval(elem_string, globals(), CLASS_MAP)


def element_from_m(line: str) -> Element:
    """Builds an :py:class:`.Element` from a line in an m-file

    Parameters:
        line:           Matlab string representation of an :py:class:`.Element`

    Returns:
        elem (Element): new :py:class:`.Element`
    """
    def argsplit(value):
        return [a.strip() for a in split_ignoring_parentheses(value, ',')]

    def makedir(mat_struct):
        """Build directory from Matlab struct arguments"""
        def pairs(it):
            while True:
                try:
                    a = next(it)
                except StopIteration:
                    break
                yield eval(a), convert(next(it))
        return dict(pairs(iter(mat_struct)))

    def makearray(mat_arr):
        """Build numpy array for Matlab array syntax"""
        def arraystr(arr):
            lns = arr.split(';')
            rr = [arraystr(v) for v in lns] if len(lns) > 1 else lns[0].split()
            return '[{0}]'.format(', '.join(rr))
        return eval('numpy.array({0})'.format(arraystr(mat_arr)))

    def convert(value):
        """convert Matlab syntax to numpy syntax"""
        if value.startswith('['):
            result = makearray(value[1:-1])
        elif value.startswith('struct'):
            result = makedir(argsplit(value[7:-1]))
        else:
            result = eval(value)
        return result

    left = line.index('(')
    right = line.rindex(')')
    matcls = line[:left].strip()[2:]
    cls = _CLASS_MAP[matcls]
    arguments = argsplit(line[left + 1:right])
    ll = len(cls._BUILD_ATTRIBUTES)
    if ll < len(arguments) and arguments[ll].endswith("Pass'"):
        arguments.insert(ll, "'PassMethod'")
    args = [convert(v) for v in arguments[:ll]]
    kwargs = makedir(arguments[ll:])
    if matcls == 'rbend':
        # the Matlab 'rbend' has no equivalent in PyAT. This adds parameters
        # necessary for using the python sector bend
        halfangle = 0.5 * args[2]
        kwargs.setdefault('EntranceAngle', halfangle)
        kwargs.setdefault('ExitAngle', halfangle)
    return cls(*args, **kwargs)


def element_to_dict(elem: Element) -> dict:
    """Builds the Matlab structure of an :py:class:`.Element`

    Parameters:
        elem:           :py:class:`.Element`

    Returns:
        dct (dict):     Dictionary of :py:class:`.Element` attributes
    """
    dct = dict((k, _mattype_map.get(type(v), lambda attr: attr)(v))
               for k, v in elem.items())
    class_name = elem.__class__.__name__
    dct['Class'] = _matclass_map.get(class_name, class_name)
    return dct


def element_to_m(elem: Element) -> str:
    """Builds the Matlab-evaluable string for an :py:class:`.Element`

    Parameters:
        elem:           :py:class:`.Element`

    Returns:
        mstr (str):     Matlab string representation of the
          :py:class:`.Element` attributes
    """

    def convert(arg):
        def convert_dict(pdir):
            def scan(d):
                for k, v in d.items():
                    yield convert(k)
                    yield convert(v)
            return 'struct({0})'.format(', '.join(scan(pdir)))

        def convert_array(arr):
            if arr.ndim > 1:
                lns = (str(list(ln)).replace(',', '')[1:-1] for ln in arr)
                return ''.join(('[', '; '.join(lns), ']'))
            elif arr.ndim > 0:
                return str(list(arr)).replace(',', '')
            else:
                return str(arr)

        if isinstance(arg, numpy.ndarray):
            return convert_array(arg)
        elif isinstance(arg, dict):
            return convert_dict(arg)
        elif isinstance(arg, Particle):
            return convert_dict(arg.to_dict())
        else:
            return repr(arg)

    def m_name(elclass):
        stdname = ''.join(('at', elclass.__name__.lower()))
        return _class_to_matfunc.get(elclass, stdname)

    attrs = dict(elem.items())
    # noinspection PyProtectedMember
    args = [attrs.pop(k, getattr(elem, k)) for k in elem._BUILD_ATTRIBUTES]
    defelem = elem.__class__(*args)
    kwds = dict((k, v) for k, v in attrs.items()
                if not numpy.array_equal(v, getattr(defelem, k, None)))
    argstrs = [convert(arg) for arg in args]
    if 'PassMethod' in kwds:
        argstrs.append(convert(kwds.pop('PassMethod')))
    argstrs += [', '.join((repr(k), convert(v))) for k, v in kwds.items()]
    return '{0:>15}({1});...'.format(m_name(elem.__class__),
                                     ', '.join(argstrs))


# Kept for compatibility but should be deprecated:


CLASS_MAPPING = dict((key, cls.__name__) for (key, cls) in _CLASS_MAP.items())

PASS_MAPPING = dict((key, cls.__name__) for (key, cls) in _PASS_MAP.items())


def find_class_name(elem_dict, quiet=False):
    """Derive the class name of an Element from its attributes"""
    return find_class(elem_dict, quiet=quiet).__name__


def split_ignoring_parentheses(string, delimiter):
    placeholder = "placeholder"
    substituted = string[:]
    matches = collections.deque(re.finditer("\\(.*?\\)", string))
    for match in matches:
        substituted = substituted.replace(match.group(), placeholder, 1)
    parts = substituted.split(delimiter)
    replaced_parts = []
    for part in parts:
        if placeholder in part:
            next_match = matches.popleft()
            part = part.replace(placeholder, next_match.group(), 1)
        replaced_parts.append(part)
    assert not matches

    return replaced_parts



import numpy
import at
import matplotlib.pyplot as plt
from matplotlib import cm
from at.collective import Wake, LongResonatorElement
from at.collective.haissinski import Haissinski

# First we define the ring, the BB resonator, the current and the wake element
ring = at.load_m('../../../machine_data/esrf.m')

freq = 10e9
qfactor = 1
Rs = 1e4
current = 5e-4
m = 50 #30 is quite coarse, 70 or 80 is very fine. 50 is middle
kmax = 8

srange = Wake.build_srange(-0.36, 0.36, 1.0e-5, 1.0e-2, ring.circumference, ring.circumference)

# Now we initialise the Haissinski class, and solve, then we normalise the distribution and shift the charge center to be at 0
wobj = Wake.long_resonator(srange, freq, qfactor, Rs, ring.beta)
ha = Haissinski(wobj, ring, m=m, kmax=kmax, current=current, numIters = 30, eps=1e-13)
ha.solve()

ha_x_tmp = ha.q_array*ha.sigma_l
ha_prof = ha.res/ha.Ic
ha_prof /= numpy.trapz(ha_prof, x=ha_x_tmp)
ha_cc = numpy.average(ha_x_tmp, weights=ha_prof)
ha_x = (ha_x_tmp - ha_cc) 

'''
currents = numpy.arange(0, 1.1e-3, 2e-4)
ha.solve_steps(currents)

fig = plt.figure()
ax1 = fig.add_subplot(111)
for i in numpy.arange(len(currents)):
    ax1.plot(1e3*ha.q_array*ha.sigma_l, ha.res_steps[i,:]/ha.I_steps[i], label='Ib={:f}mA'.format(currents[i]*1e3), color=cm.jet(float(i)/len(currents)))
ax1.legend()
ax1.set_xlabel('z [mm]')
ax1.set_ylabel(r'$\rho(z)$')
plt.show()

'''

# Now we set up and run the tracking. The final distribution is an average of the last numAve turns 
welem = LongResonatorElement('wake', ring, srange, freq, qfactor, Rs, Nslice=300)
welem.Current = current


ring.radiation_on()
ring.set_cavity_phase()
_,fring = at.fast_ring(ring)
fring.append(welem)


sigm = at.sigma_matrix(ring.radiation_on(copy=True))
Nparts = 10000
Nturns = 20000
nbins = 60
part = at.beam(Nparts, sigm)

histAveFlag = False
numAve = 5000
id0 = 0
for t in numpy.arange(Nturns):
    if t%1000==0:
        print('Tracking turn ', t, ' of ', Nturns)
    part = at.lattice_pass(fring, part)[:,:,0,0]
    
    if t > Nturns-numAve:  
        if not histAveFlag:
            histAveFlag = True

            nt1, et1 = numpy.histogram(part[5,:], bins=nbins)

            etc1 = numpy.array([(et1[i] + et1[i+1])/2 for i in numpy.arange(len(et1)-1)]) 

            allData = numpy.zeros((len(etc1), numAve + 1))
            allData[:, 0] = etc1
            allData[:, id0 + 1] =  nt1

        else:
            n1, e1 = numpy.histogram(part[5,:], bins=et1)
            ec1 = numpy.array([(e1[i] + e1[i+1])/2 for i in numpy.arange(len(e1)-1)]) 
            allData[:, id0 + 1] =  n1
        id0 += 1   
        

# Identical post processing of tracking distribution (to the Haissinski)
zr_tmp = allData[:,0]
prof = numpy.mean(allData[:,1:],axis=1) 
prof /= numpy.trapz(prof,x=zr_tmp)

cc = numpy.average(zr_tmp, weights=prof)
zr = zr_tmp - cc

fig = plt.figure()
ax1 = fig.add_subplot(111)
ax1.plot(1e3*ha_x, ha_prof, color='r', linestyle='solid', label='Haissinski Solution')
ax1.plot(1e3*zr, prof, color='k', linestyle='dashed', label='Tracking')
ax1.set_xlabel('z [mm]')
ax1.set_ylabel(r'$\rho(z)$')
ax1.legend()
plt.show()






"""Global set of constants"""
import os
import multiprocessing


class _Dst(object):
    """Set of constants for AT numerical analysis"""
    # optimized values:
    # see https://github.com/atcollab/at/pull/199#issuecomment-745420443
    XYStep = 3.e-8           # Coordinate step for differentiation
    DPStep = 3.e-6           # Momentum step for dispersion and chromaticity
    OrbConvergence = 1.e-12  # Convergence criterion for orbit
    OrbMaxIter = 20          # Max. number of iterations for orbit
    omp_num_threads = int(os.environ.get('OMP_NUM_THREADS', '0'))
    patpass_poolsize = multiprocessing.cpu_count()

    def __setattr__(self, name, value):
        _ = getattr(self, name)     # make sure attribute exists
        object.__setattr__(self, name, value)

    def reset(self, name):
        delattr(self, name)


DConstant = _Dst()
"""
Default values for AT algorithms

Attributes:
    XYStep:             Coordinate step for differentiation
    DPStep:             Momentum step for dispersion and chromaticity
    OrbConvergence:     Convergence criterion for orbit
    OrbMaxIter:         Max. number of iterations for orbit
    omp_num_threads:    Default number of OpenMP threads
"""



"""
Functions used to
calculate the loss boundary for different
grid definitions
"""

from at.lattice import Lattice, AtError
from at.tracking import lattice_pass, patpass
from typing import Optional, Sequence
from enum import Enum
import numpy
from scipy.ndimage import binary_dilation, binary_opening
from collections import namedtuple
import time

__all__ = ['GridMode']

_pdict = {'x': 0, 'xp': 1,
          'y': 2, 'yp': 3,
          'dp': 4, 'ct': 5}


class GridMode(Enum):
    """
    Grid definition for 2D acceptance boundary search
    """
    RADIAL = 0      #: full [:math:`\:r, \theta\:`] grid
    CARTESIAN = 1   #: full [:math:`\:x, y\:`] grid
    RECURSIVE = 2   #: radial recursive search


def grid_config(planes, amplitudes, npoints, bounds, grid_mode,
                shift_zero):
    """"
    Returns an object that defines the grid configuration
    """
    bounds = numpy.atleast_2d(bounds)
    bounds = tuple(map(tuple, bounds))
    shape = numpy.array(npoints, dtype=numpy.int32)
    d = {'planes': numpy.atleast_1d(planes),
         'planesi': numpy.atleast_1d(get_plane_index(planes)),
         'amplitudes': numpy.atleast_1d(amplitudes),
         'shape': numpy.atleast_1d(shape),
         'bounds': bounds,
         'mode': grid_mode,
         'shift_zero': shift_zero}
    return namedtuple('config', d.keys())(**d)


def grid(grid, offset):
    """"
    Returns a grid object
    """
    d = {'grid': numpy.atleast_2d(grid),
         'offset': numpy.atleast_1d(offset)}
    return namedtuple('grid', d.keys())(**d)


def get_plane_index(planes):
    """
    Converts plane to particle coordinate index
    """
    planesi = numpy.array([], dtype=numpy.int32)
    for i, p in enumerate(numpy.atleast_1d(planes)):
        if isinstance(p, str):
            try:
                planesi = numpy.append(planesi, _pdict[p])
            except KeyError:
                raise AtError('Allowed values for plane are x,xp,y,yp,dp,ct')
        else:
            raise AtError('Allowed values for plane are x,xp,y,yp,dp,ct')
    return planesi


def set_ring_orbit(ring, dp, obspt, orbit):
    """
    Returns a ring starting at obspt and initial
    closed orbit
    """
    if obspt is not None:
        assert numpy.isscalar(obspt), 'Scalar value needed for obspt'
        ring = ring.rotate(obspt)

    if orbit is None:
        orbit = ring.find_orbit(dp=dp)[0]

    return orbit, ring


def grid_configuration(planes, npoints, amplitudes, grid_mode, bounds=None,
                       shift_zero=1.0e-9):
    """
    Return a grid configuration based on user input parameters, the ordering
    is as follows: CARTESIAN: (x,y), RADIAL/RECURSIVE (r, theta).
    Scalar inputs can be used for 1D grid
    """
    ndims = len(numpy.atleast_1d(planes))
    if ndims > 2 or ndims == 0:
        raise AtError('planes can have 1 or 2 element (1D or 2D aperture)')
    elif ndims == 1 and grid_mode is GridMode.RADIAL:
        grid_mode = GridMode.CARTESIAN

    if numpy.shape(numpy.atleast_1d(npoints)) != (ndims,):
        raise AtError('npoints shape should be (len(planes),)')
    if numpy.shape(numpy.atleast_1d(amplitudes)) != (ndims,):
        raise AtError('amplitudes shape should be (len(planes),)')
    if (numpy.shape(numpy.atleast_2d(bounds)) != (ndims, 2)
       and bounds is not None):
        raise AtError('bounds shape should be (len(planes),2)')

    if grid_mode is GridMode.RADIAL or grid_mode is GridMode.RECURSIVE:
        if bounds is None:
            bounds = numpy.array([[0, 1], [numpy.pi, 0]])
        bounds[0][bounds[0] == 0] = 1.0e-6
    elif grid_mode is GridMode.CARTESIAN:
        if bounds is None:
            bounds = numpy.array([[p-1, 1] for p in range(ndims)])
    else:
        raise AtError('GridMode {0} undefined.'.format(grid_mode))

    config = grid_config(planes, amplitudes, npoints,
                         bounds, grid_mode, shift_zero)
    return config


def get_parts(config, offset):
    """
    Generate a (6,n) particles array based on the grid configuration
    and initial offset, returns a grid object containing the grid and
    the offset from which the array can be reconstructed
    """
    def get_part_grid_uniform(bnd, np, amp):
        x = [numpy.linspace(*b, n)*a for a, b, n in zip(amp, bnd, np)]
        try:
            g1, g2 = numpy.meshgrid(*x)
        except ValueError:
            g1, g2 = numpy.meshgrid(x, 0.0)
        return numpy.array([g1.flatten(), g2.flatten()])

    def get_part_grid_radial(bnd, np, amp):
        x = [numpy.linspace(*b, n) for b, n in zip(bnd, np)]
        g1, g2 = numpy.meshgrid(*x)
        g1r = amp[0]*numpy.cos(g2)*g1
        g2r = amp[1]*numpy.sin(g2)*g1
        return numpy.array([g1r.flatten(), g2r.flatten()])

    pind = config.planesi
    amp = config.amplitudes
    np = config.shape
    bnd = config.bounds
    gm = config.mode

    if gm is GridMode.CARTESIAN:
        g = get_part_grid_uniform(bnd, np, amp)
    elif gm is GridMode.RADIAL:
        g = get_part_grid_radial(bnd, np, amp)
    parts = numpy.zeros((6, numpy.prod(np)))
    parts[pind, :] = [g[i] for i in range(len(pind))]
    if len(pind) == 2:
        parts[pind[0]][parts[pind[1]] == 0.0] += config.shift_zero
        parts[pind[1]][parts[pind[0]] == 0.0] += config.shift_zero
    parts = (parts.T+offset).T
    return parts, grid(g, offset[pind])


def get_survived(parts, ring, nturns, use_mp, **kwargs):
    """
    Track a grid through the ring and extract survived particles
    """
    if use_mp:
        _ = patpass(ring, parts, nturns=nturns, **kwargs)
    else:
        _ = lattice_pass(ring, parts, nturns=nturns, **kwargs)
    if parts.ndim == 1:
        survived = numpy.invert(numpy.isnan(parts[0]))
    else:
        survived = numpy.invert(numpy.isnan(parts[0, :]))
    return survived


def get_grid_boundary(mask, grid, config):
    """
    Compute the boundary of survided particles array
    """
    def nearest_order(grid):
        #  keep only max r for each angle
        angle = numpy.arctan2(*grid)
        norm = numpy.linalg.norm(grid.T, axis=1)
        val, inv = numpy.unique(angle, return_inverse=True)
        gf = numpy.zeros((2, len(val)))
        for i, v in enumerate(val):
            inds = numpy.where(inv == i)[0]
            ni = norm[inds]
            nim = numpy.where(ni == numpy.amax(ni))[0]
            ind = inds[nim][0]
            gf[:, i] = grid[:, ind]
        #  first sort by angle
        idx = numpy.argsort(numpy.arctan2(*gf))
        gf = gf[:, idx]
        #  now sort by closest neighbour on normalized grid
        x, y = gf[0, :].copy(), gf[1, :].copy()
        dxmin = min(numpy.diff(numpy.unique(x)))
        dymin = min(numpy.diff(numpy.unique(y)))
        iorder = [0]
        for i in range(1, len(x)):
            xnow = x[iorder[-1]]
            ynow = y[iorder[-1]]
            dd = numpy.sqrt(((x-xnow)/dxmin)**2+((y-ynow)/dymin)**2)
            if i <= 3:
                ic = [j for j in numpy.argsort(dd) if j not in iorder]
            else:
                direction = numpy.sign(iorder[-1]-iorder[-2])
                ic = [j for j in numpy.argsort(dd) if j not in iorder
                      and numpy.sign(j-iorder[-1]) == direction]
            if len(ic) > 0:
                iorder.append(ic[0])
        #  finally connect both ends if distance within unit square
        xnow = x[iorder[-1]]
        ynow = y[iorder[-1]]
        xs = x[iorder[0]]
        ys = y[iorder[0]]
        dd = numpy.sqrt(((xs-xnow)/dxmin)**2+((ys-ynow)/dymin)**2)
        if dd < 1.5:
            iorder.append(iorder[0])
        gf = gf[:, iorder]
        return gf

    def search_bnd(ma, sa):
        bnd = numpy.zeros((2, 1))
        for j, m in enumerate(ma):
            bnd = sa[:, j]
            if not m and j > 0:
                bnd = sa[:, j-1]
                break
        return bnd

    def grid_boundary(mask, grid, config, nclean=2):
        cnt = numpy.flip(config.shape)
        if len(cnt) == 1:
            return vector_boundary(mask, grid)
        mask2d = numpy.reshape(mask.copy(), cnt)
        #  remove isolated points
        for i in range(nclean):
            bnd1 = binary_opening(mask2d, numpy.ones((2, 1)))
            bnd2 = binary_opening(mask2d, numpy.ones((1, 2)))
        bnd = numpy.logical_and(bnd1, bnd2)
        k = numpy.zeros((3, 3), dtype=int)
        k[1] = 1
        k[:, 1] = 1
        bnd = numpy.logical_and(binary_dilation(bnd == 0, border_value=1), bnd)
        bnd = grid.grid[:, bnd.reshape(mask.shape)]
        return nearest_order(bnd)

    def vector_boundary(mask, grid):
        g = grid.grid
        xp, xn = g[0] >= 0, g[0] <= 0
        bp = search_bnd(mask[xp], g[:, xp])
        bn = search_bnd(numpy.flip(mask[xn]),
                        numpy.flip(g[:, xn], axis=1))
        return numpy.squeeze([bn[0], bp[0]])

    def radial_boundary(mask, grid):
        angles = numpy.round(numpy.arctan2(*grid.grid), decimals=8)
        angles, invi = numpy.unique(angles, return_inverse=True)
        bnd = numpy.zeros((2, len(angles)))
        for i in range(len(angles)):
            sa = grid.grid[:, invi == i]
            ma = mask[invi == i]
            bnd[:, i] = search_bnd(ma, sa)
        return bnd

    if not numpy.any(mask):
        raise AtError("No particle survived, please check your grid "
                      "or lattice.")

    if config.mode is GridMode.RADIAL:
        return radial_boundary(mask, grid)
    elif config.mode is GridMode.CARTESIAN:
        return grid_boundary(mask, grid, config)
    else:
        raise AtError('GridMode {0} undefined.'.format(grid.mode))


def grid_boundary_search(ring, planes, npoints, amplitudes, nturns=1024,
                         obspt=None, dp=None, offset=None, bounds=None,
                         grid_mode=GridMode.RADIAL, use_mp=False,
                         verbose=True, shift_zero=1.0e-9, **kwargs):
    """
    Search for the boundary by tracking a grid
    """
    offset, newring = set_ring_orbit(ring, dp, obspt,
                                     offset)
    config = grid_configuration(planes, npoints, amplitudes,
                                grid_mode, bounds=bounds,
                                shift_zero=shift_zero)

    if verbose:
        print('\nRunning grid boundary search:')
        if obspt is None:
            print('Element {0}, obspt={1}'.format(ring[0].FamName, 0))
        else:
            print('Element {0}, obspt={1}'.format(ring[obspt].FamName,
                                                  obspt))
        print('The grid mode is {0}'.format(config.mode))
        print('The planes are {0}'.format(config.planes))
        print('Number of steps are {0}'.format(config.shape))
        print('The maximum amplitudes are {0}'.format(config.amplitudes))
        print('The maximum boundaries are {0}'.format(config.bounds))
        print('The initial offset is {0} with dp={1}'.format(offset, dp))

    t0 = time.time()
    parts, grid = get_parts(config, offset)
    mask = get_survived(parts, newring, nturns, use_mp, **kwargs)
    survived = grid.grid[:, mask]
    boundary = get_grid_boundary(mask, grid, config)
    if verbose:
        print('Calculation took {0}'.format(time.time()-t0))
    return boundary, survived, grid.grid


def recursive_boundary_search(ring, planes, npoints, amplitudes, nturns=1024,
                              obspt=None, dp=None, offset=None, bounds=None,
                              use_mp=False, divider=2, verbose=True,
                              shift_zero=1.0e-9, **kwargs):
    """
    Recursively search for the boundary in a given plane and direction (angle)
    """
    def search_boundary(planesi, angles, rtol, rsteps, nturns,
                        offset, use_mp, **kwargs):

        ftol = min(rtol/rsteps)
        cs = numpy.squeeze([numpy.cos(angles), numpy.sin(angles)])
        cs = numpy.around(cs, decimals=9)
        fact = numpy.ones(len(angles))
        survived = numpy.full(len(angles), True)
        istracked = numpy.full(len(angles), True)
        part = numpy.zeros((6, len(angles)))
        grid = numpy.array([])
        mask = numpy.array([])

        while numpy.any(survived):
            for i, pi in enumerate(planesi):
                part[pi, survived] += cs[i, survived]*rsteps[i]*fact[survived]
            istracked = numpy.array([not numpy.any([numpy.allclose(p, g,
                                                                   rtol=1.0e-9)
                                    for g in grid.T])
                                    for p in part[planesi].T])
            survived = numpy.array([numpy.any([numpy.allclose(p, m,
                                                              rtol=1.0e-9)
                                   for m in mask.T])
                                   for p in part[planesi].T])
            pt = part[:, istracked]
            grid = numpy.hstack([grid,
                                 pt[planesi]]) if grid.size else pt[planesi]
            ptmp = (pt.T + offset).T
            survived[istracked] = get_survived(ptmp, newring, nturns,
                                               use_mp, **kwargs)
            pm = part[:, numpy.logical_and(istracked, survived)]
            mask = numpy.hstack([mask,
                                 pm[planesi]]) if mask.size else pm[planesi]
            for i in range(len(angles)):
                if not survived[i] and fact[i] > ftol:
                    deltas = cs[:, i]*rsteps[:]*min(1, 2*fact[i])
                    if numpy.any(abs(deltas) > abs(part[planesi, i])):
                        part[planesi, i] = numpy.zeros(len(planesi))
                    else:
                        for j, pi in enumerate(planesi):
                            part[pi, i] -= deltas[j]
                    survived[i] = True
                    fact[i] *= 1/divider

        for i, pi in enumerate(planesi):
            part[pi] -= cs[i]*rsteps[i]*fact

        p = numpy.squeeze(part[planesi])
        return p, mask, grid

    offset, newring = set_ring_orbit(ring, dp, obspt, offset)
    config = grid_configuration(planes, npoints, amplitudes,
                                GridMode.RECURSIVE, bounds=bounds,
                                shift_zero=shift_zero)
    rtol = min(numpy.atleast_1d(config.amplitudes/config.shape))
    rstep = config.amplitudes
    if len(numpy.atleast_1d(config.shape)) == 2:
        angles = numpy.linspace(*config.bounds[1], config.shape[1])
    else:
        angles = numpy.linspace(*config.bounds[1], 2)
    angles = numpy.atleast_1d(angles)

    if verbose:
        print('\nRunning recursive boundary search:')
        if obspt is None:
            print('Element {0}, obspt={1}'.format(ring[0].FamName, 0))
        else:
            print('Element {0}, obspt={1}'.format(ring[obspt].FamName,
                                                  obspt))
        print('The grid mode is {0}'.format(config.mode))
        print('The planes are {0}'.format(config.planes))
        print('Number of angles is {0} from {1} to {2} rad'.format(len(angles),
              angles[0], angles[-1]))
        print('The resolution of the search is {0}'.format(rtol))
        print('The initial step size is {0}'.format(rstep))
        print('The initial offset is {0} with dp={1}'.format(offset, dp))

    t0 = time.time()
    result = search_boundary(config.planesi, angles, rtol, rstep,
                             nturns, offset, use_mp, **kwargs)
    if verbose:
        print('Calculation took {0}'.format(time.time()-t0))
    return result


def boundary_search(ring: Lattice, planes, npoints, amplitudes,
                    nturns: Optional[int] = 1024,
                    obspt: Optional[int] = None, dp: Optional[float] = None,
                    offset: Sequence[float] = None, bounds=None,
                    grid_mode: Optional[GridMode] = GridMode.RADIAL,
                    use_mp: Optional[bool] = False,
                    verbose: Optional[bool] = True,
                    shift_zero: Optional[float] = 1.0e-9,
                    **kwargs):
    """
    Computes the loss boundary at a single point in the machine
    """
    divider = kwargs.pop('divider', 2)
    if grid_mode is GridMode.RECURSIVE:
        result = recursive_boundary_search(ring, planes, npoints, amplitudes,
                                           nturns=nturns, obspt=obspt, dp=dp,
                                           offset=offset, bounds=bounds,
                                           use_mp=use_mp, verbose=verbose,
                                           divider=divider,
                                           shift_zero=shift_zero,
                                           **kwargs)
    else:
        result = grid_boundary_search(ring, planes, npoints, amplitudes,
                                      nturns=nturns, obspt=obspt, dp=dp,
                                      offset=offset, bounds=bounds,
                                      grid_mode=grid_mode, use_mp=use_mp,
                                      verbose=verbose, shift_zero=shift_zero,
                                      **kwargs)
    return result



from at import physics, get_s_pos, uint32_refpts
from scipy.constants import speed_of_light
import matlab
import numpy
from numpy.testing import assert_allclose as assert_close
import pytest


def _py_data(ml_data):
    """Convert a Matlab vector to a numpy vector"""
    return numpy.squeeze(numpy.asarray(ml_data))


def _ml_refs(refpts, nelems):
    """Convert refpoints to Matlab"""
    uintrefs = uint32_refpts(refpts, nelems)
    return matlab.double([float(ref+1) for ref in uintrefs])


def _compare_physdata(py_data, ml_data, fields, atol=12, rtol=1.e-7):
    for (ml_key, py_key) in fields:
        ml_val = _py_data(ml_data[ml_key])
        if py_key == 'closed_orbit':
            py_val = numpy.squeeze(py_data[py_key][:, :4])
        else:
            py_val = numpy.squeeze(py_data[py_key])
        assert_close(py_val, ml_val, atol=atol, rtol=rtol)


@pytest.mark.parametrize('dp', (-0.01, 0.0, 0.01))
@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba'),
                          pytest.lazy_fixture('dba')])
def test_linear_analysis(engine, lattices, dp):
    py_lattice, ml_lattice, _ = lattices
    nelems = len(py_lattice)
    fields = [('SPos', 's_pos'),
              ('ClosedOrbit', 'closed_orbit'),
              ('Dispersion', 'dispersion'),
              ('alpha', 'alpha'), ('beta', 'beta'),
              ('mu', 'mu'), ('M44', 'm44'),
              ('A', 'A'), ('B', 'B'), ('C', 'C'),
              ('gamma', 'gamma')]
    refpts = range(nelems + 1)
    py_data0, py_tune, py_chrom, py_data = physics.linopt(py_lattice, dp,
                                                          refpts, True,
                                                          ddp=1.E-6)
    # Matlab call
    ml_data, ml_tune, ml_chrom = engine.pyproxy('atlinopt', ml_lattice, dp,
                                                _ml_refs(refpts, nelems),
                                                nargout=3)
    ml_data0 = engine.pyproxy('atlinopt', ml_lattice, dp,
                              _ml_refs(nelems, nelems), nargout=3)[0]
    # Comparison
    assert_close(py_tune, _py_data(ml_tune), atol=1.e-8, rtol=0)
    assert_close(py_chrom, _py_data(ml_chrom), atol=3.e-3, rtol=0)
    _compare_physdata(numpy.expand_dims(py_data0, 0), ml_data0, fields,
                      rtol=1e-8)
    _compare_physdata(py_data, ml_data, fields, rtol=1e-8)


@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba')])
def test_ohmi_envelope(engine, lattices):
    py_lattice, ml_lattice, _ = lattices
    fields = [('beam66', 'r66'), ('beam44', 'r44'), ('emit66', 'emitXYZ'),
              ('emit44', 'emitXY')]
    nelems = len(py_lattice)
    refpts = range(nelems + 1)

    # Python call
    py_lattice = py_lattice.radiation_on(copy=True)
    py_emit0, py_beamdata, py_emit = py_lattice.ohmi_envelope(refpts)
    # Matlab call
    ml_emit = engine.pyproxy('atx', ml_lattice, 0.0, _ml_refs(refpts, nelems))
    ml_emit0, ml_params = engine.pyproxy('atx', ml_lattice, 0.0,
                                         _ml_refs(0, nelems), nargout=2)
    revolution_period = get_s_pos(py_lattice, nelems) / speed_of_light
    damping_times = revolution_period / py_beamdata.damping_rates
    # Comparison
    assert_close(damping_times, _py_data(ml_params['dampingtime']),
                 atol=1.e-6, rtol=1.e-3)
    assert_close(py_beamdata.mode_emittances, _py_data(ml_emit0['modemit']),
                 atol=1.e-20, rtol=2.e-3)
    _compare_physdata(numpy.expand_dims(py_emit0, 0), ml_emit0, fields)
    _compare_physdata(py_emit, ml_emit, fields)


@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba_rad')])
def test_quantdiff(engine, lattices):
    py_lattice, ml_lattice, radindex = lattices
    # Python call
    dmat = physics.radiation.quantdiffmat(py_lattice)
    # Matlab call
    dmat_ml = engine.pyproxy('quantumDiff', ml_lattice, radindex)
    # Comparison
    assert_close(dmat, dmat_ml, rtol=1.0e-8, atol=1.0e-18)


@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba')])
def test_fastring(engine, lattices):
    py_lattice, ml_lattice, _ = lattices
    # Python call
    ring, ringrad = physics.fastring.fast_ring(py_lattice)
    # Matlab call
    ring_ml, ringrad_ml = engine.pyproxy('atfastring',
                                         ml_lattice, nargout=2)
    # Comparison
    for r, rml in zip([ring, ringrad], [ring_ml, ringrad_ml]):
        assert_close(r[0].Frequency,    rml[1]['Frequency'], rtol=1.0e-20)
        assert_close(r[0].Voltage,      rml[1]['Voltage'], rtol=1.0e-20)
        assert_close(r[1].I2,           rml[2]['I2'], rtol=1.0e-20)
        assert_close(r[1].Length,       rml[2]['Length'], rtol=1.0e-20)
        assert_close(r[1].M66,          rml[2]['M66'], atol=1.0e-7)
        assert_close(r[1].T1,   numpy.squeeze(rml[2]['T1']),
                     rtol=1.0e-8, atol=1.0e-11)
        assert_close(r[1].T2,   numpy.squeeze(rml[2]['T2']),
                     rtol=1.0e-8, atol=1.0e-11)
        assert_close(r[2].A1,       rml[-1]['A1'], rtol=0.02)
        assert_close(r[2].A2,       rml[-1]['A2'], rtol=0.02)
        assert_close(r[2].A3,       rml[-1]['A3'], rtol=0.02)
        assert_close(r[2].Alphax,   rml[-1]['Alphax'], rtol=1.0e-5)
        assert_close(r[2].Alphay,   rml[-1]['Alphay'], rtol=1.0e-6)
        assert_close(r[2].Betax,    rml[-1]['Betax'], rtol=1.0e-10)
        assert_close(r[2].Betay,    rml[-1]['Betay'], rtol=1.0e-10)
        assert_close(r[2].Qpx,      rml[-1]['Qpx'], rtol=1.0e-5)
        assert_close(r[2].Qpy,      rml[-1]['Qpy'], rtol=1.0e-5)
        assert_close(r[2].T1,   numpy.squeeze(rml[-1]['T1']),
                     rtol=1.0e-8, atol=1.0e-11)
        assert_close(r[2].T2,   numpy.squeeze(rml[-1]['T2']),
                     rtol=1.0e-8, atol=1.0e-11)
        if len(r) >= 4:
            assert_close(r[3].Lmatp,    rml[-2]['Lmatp'],
                         rtol=0.02, atol=1.0e-10)


@pytest.mark.parametrize('dp', (0.00, 0.01, -0.01))
@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba')])
def test_parameters(engine, lattices, dp):
    py_lattice, ml_lattice, _ = lattices

    # Test perimeter
    py_length = py_lattice.get_s_pos(len(py_lattice))
    ml_length = engine.findspos(ml_lattice, len(ml_lattice)+1)
    assert_close(py_length, ml_length, rtol=1.E-8)

    # test energy loss
    ml_energy, ml_periods, ml_voltage, \
        ml_harms, ml_eloss = engine.pyproxy('atenergy', ml_lattice, nargout=5)
    assert_close(py_lattice.rf_voltage, ml_voltage, rtol=1.E-8)
    assert_close(py_lattice.energy_loss, ml_eloss, rtol=1.E-6)
    assert py_lattice.energy == ml_energy
    assert py_lattice.periodicity == int(ml_periods)
    assert py_lattice.harmonic_number == int(ml_harms)

    # test momentum compaction factor
    py_mcf = py_lattice.get_mcf(dp)
    ml_mcf = engine.mcf(ml_lattice, dp)
    assert_close(py_mcf, ml_mcf, rtol=1.E-8)



"""
Functions relating to particle generation
"""
import numpy
from at.physics import ohmi_envelope, radiation_parameters
from at.constants import clight
from at.lattice import AtError

__all__ = ['beam', 'sigma_matrix']


def _generate_2d_trans_matrix(emit, beta, alpha):
    return emit*numpy.array([[beta, -alpha], [-alpha, (1+alpha**2)/beta]])


def _generate_2d_long_matrix(espread, blength):
    return numpy.array([[espread*espread, 0], [0, blength*blength]])


def _generate_2d_long_Rmatrix(espread, blength):
    return numpy.array([[espread/blength, 0], [0, blength/espread]])


def _sigma_matrix_uncoupled(betax, alphax, emitx, betay, alphay, emity,
                            blength, espread):
    sig_matrix = numpy.zeros((6, 6))
    sig_matrix[:2, :2] = _generate_2d_trans_matrix(emitx, betax, alphax)
    sig_matrix[2:4, 2:4] = _generate_2d_trans_matrix(emity, betay, alphay)
    sig_matrix[4:, 4:] = _generate_2d_long_matrix(espread, blength)
    return sig_matrix


def _sigma_matrix_from_R66(R66, emitx, emity, blength, espread):
    sig_matrix = emitx*R66[0] + emity*R66[1] + \
             blength*espread*R66[2]
    return sig_matrix


def _compute_bunch_length_from_espread(ring, espread):
    rp = radiation_parameters(ring.radiation_off(copy=True))
    blength = ring.beta*clight*numpy.abs(rp.etac)*espread / \
        (2.0*numpy.pi*rp.f_s)
    return blength


def _sigma_matrix_lattice(ring=None, twiss_in=None, emitx=None, emity=None,
                          blength=None, espread=None, verbose=False):

    if ring is None and (emitx is None or emity is None or blength is None
                         or espread is None or twiss_in is None):
        raise AtError('Provide either a ring or twiss_in and ALL emittance '
                      'paramaters: (ex, ey, espread, blength)')

    if espread is not None and blength is None:
        blength = _compute_bunch_length_from_espread(ring, espread)
    flag_all = emitx and emity and espread
    flag_any = emitx or emity or espread

    if not flag_all:
        if verbose:
            print('Calculating missing parameters (ex, ey or espread) '
                  'using ohmi envelop: needs radiations ON')
        try:
            emit0, beamdata, emit = ohmi_envelope(ring, refpts=[0])
        except AtError:
            raise AtError('Please provide ex, ey, espread or turn on '
                          'radiations to compute the sigma matrix')
        if emitx is None:
            emitx = beamdata.mode_emittances[0]
        if emity is None:
            emity = beamdata.mode_emittances[1]
        if espread is None:
            espread = numpy.sqrt(emit0.r66[4, 4])
            blength = numpy.sqrt(emit0.r66[5, 5])

    if not flag_any and not twiss_in:
        return emit.r66[0]
    elif twiss_in:
        if verbose:
            print('Generating pseudo-correlated matrix '
                  'from twiss_in')
        if not hasattr(twiss_in, 'R'):
            raise AtError('twiss_in has to the contain the R matrix. '
                          'Please use the output from linopt6.')
        rmat = twiss_in.R
    else:
        if verbose:
            print('Generating pseudo-correlated matrix '
                  'from start point of ring')
        l0, _, _ = ring.get_optics()
        rmat = l0.R

    if rmat.shape[0] != 3:
        rmat6 = numpy.zeros((3, 6, 6))
        rmat6[0, :4, :4] = rmat[0]
        rmat6[1, :4, :4] = rmat[1]
        rmat6[2, 4:, 4:] = _generate_2d_long_Rmatrix(espread, blength)
    else:
        rmat6 = rmat
    sig_matrix = _sigma_matrix_from_R66(rmat6, emitx, emity, blength,
                                        espread)
    return sig_matrix


def sigma_matrix(ring=None, twiss_in=None, betax=None, alphax=None,
                 emitx=None, betay=None, alphay=None, emity=None,
                 blength=None, espread=None, verbose=False):
    """
    Calculate the correlation matrix to be used for particle generation

    Parameters:
        ring:           Lattice object or list of twiss parameters.
        twiss_in:       Data structure containing input twiss parameters.
        betax:          Input horizontal beta function [m]
        alphax:         Input horizontal alpha function [m]
        emitx:          Horizontal emittance [m.rad]
        betay:          Input vertical beta function [m]
        alphay:         Input vertical alpha function [m]
        emity:          Vertical emittance [m.rad]
        blength:        One sigma bunch length [m]
        espread:        One sigma energy spread [dp/p]
        verbose:        Boolean flag on whether to print information
                        to the terminal
    Returns:
        sigma_matrix:    6x6 correlation matrix

    If the lattice object is provided ``ohmi_envelope`` is used to
    compute the correlated sigma matrix and missing emittances
    and longitudinal parameters

    If emittances and longitudinal parameters are provided, then
    the 2x2 uncorrelated matrices are computed for each plane (x,y,z)
    using the initial optics computed from ring.get_optics,
    and are combined together into the 6x6 matrix.

    If ``twiss_in`` is provided, it has to be produced with ``linopt6`` and
    contain the rmatrix. ``ring`` is used only to compute missing emittances.

    If neither a lattice object nor a twiss_in is provided,
    then the ``beta``, ``alpha`` and ``emittance`` for horizontal and
    vertical are required, as well as ``blength`` and ``espread``.
    This then computes the analytical uncoupled sigma matrix
    """
    if ring is not None or twiss_in is not None:
        return _sigma_matrix_lattice(ring=ring, twiss_in=twiss_in,
                                     emitx=emitx, emity=emity,
                                     blength=blength, espread=espread,
                                     verbose=verbose)
    else:
        ln = '{:s} must be defined for the uncoupled sigma_matrix'
        assert betax is not None, ln.format('betax')
        assert alphax is not None, ln.format('alphax')
        assert emitx is not None, ln.format('emitx')
        assert betay is not None, ln.format('betay')
        assert alphay is not None, ln.format('alphay')
        assert emity is not None, ln.format('emity')
        assert blength is not None, ln.format('blength')
        assert espread is not None, ln.format('espread')
        return _sigma_matrix_uncoupled(betax, alphax, emitx,
                                       betay, alphay, emity,
                                       blength, espread)


def beam(nparts: int, sigma, orbit=None):
    """
    Generates an array of random particles according to the given sigma
    matrix

    Parameters:
        nparts:         Number of particles
        sigma:          sigma_matrix as calculated by at.sigma_matrix
        orbit:          An orbit can be provided to give a center of
                        mass offset to the distribution
    Returns:
        particle_dist:  a (6, ``nparts``) matrix of coordinates
    """
    def _get_single_plane(dims):
        row_idx = numpy.array(dims)
        try:
            return numpy.linalg.cholesky(sigma[row_idx[:, None], row_idx])
        except numpy.linalg.LinAlgError:
            return numpy.zeros((2, 2))

    v = numpy.random.normal(size=(sigma.shape[0], nparts))

    try:
        lmat = numpy.linalg.cholesky(sigma)
    except numpy.linalg.LinAlgError:
        lmat = numpy.zeros((6, 6))
        lmat[:2, :2] = _get_single_plane([0, 1])
        lmat[2:4, 2:4] = _get_single_plane([2, 3])
        lmat[4:, 4:] = _get_single_plane([4, 5])

    particle_dist = numpy.squeeze(numpy.dot(lmat, v))
    if particle_dist.ndim == 1:
        particle_dist = numpy.array([particle_dist]).T

    if orbit is not None:
        if numpy.shape(orbit) != (6,):
            raise AtError('beam: input orbit shape has to be (6,)')
        particle_dist = (particle_dist.T + numpy.array(orbit)).T

    return particle_dist



"""
Functions relating to fast_ring
"""
from functools import reduce
import numpy
from typing import Tuple
from at.lattice import RFCavity, Marker, Lattice, get_cells, checkname
from at.lattice import get_elements
from at.physics import gen_m66_elem, gen_detuning_elem, gen_quantdiff_elem
import copy


__all__ = ['fast_ring']


def _rearrange(ring: Lattice, split_inds=[]):
    inds = numpy.append(split_inds, [0, len(ring)+1])
    inds = numpy.unique(inds)
    all_rings = [ring[int(b):int(e)] for b, e in zip(inds[:-1], inds[1:])]

    ringm = []
    for ring_slice in all_rings:
        ring_slice.insert(0, Marker('xbeg'))
        ring_slice.append(Marker('xend'))
        cavs = [e for e in ring_slice if isinstance(e, RFCavity)]
        newpass = ['IdentityPass' if c.Length == 0
                   else 'DriftPass' for c in cavs]
        for c, pm in zip(cavs, newpass):
            c.PassMethod = pm
        uni_freq = numpy.unique([e.Frequency for e in cavs])
        for fr in numpy.atleast_1d(uni_freq):
            cavf = [c for c in cavs if c.Frequency == fr]
            vol = reduce(lambda x, y: x+y, (c.Voltage for c in cavf))
            cavl = RFCavity('CAVL', 0, vol, fr,
                            cavf[0].HarmNumber, cavf[0].Energy)
            cavl.TimeLag = cavf[0].TimeLag
            ring_slice.append(cavl)
        ringm = ringm + ring_slice
    return all_rings, Lattice(ringm, energy=ring.energy)


def _fring(ring, split_inds=[], detuning_elem=None):
    all_rings, merged_ring = _rearrange(ring, split_inds=split_inds)
    ibegs = get_cells(merged_ring, checkname('xbeg'))
    iends = get_cells(merged_ring, checkname('xend'))
    _, orbit = merged_ring.find_orbit(refpts=ibegs | iends)
    if detuning_elem is None:
        detuning_elem = gen_detuning_elem(merged_ring, orbit[-1])
    else:
        detuning_elem.T1 = -orbit[-1]
        detuning_elem.T2 = orbit[-1]

    fastring = []
    for counter, r in enumerate(all_rings):
        cavs = [e for e in r if e.PassMethod.endswith('CavityPass')]
        [r.remove(c) for c in cavs]
        lin_elem = gen_m66_elem(r, orbit[2*counter],
                                orbit[2*counter+1])
        lin_elem.FamName = lin_elem.FamName + '_' + str(counter)
        [fastring.append(c) for c in cavs]
        fastring.append(lin_elem)
    fastring.append(detuning_elem)
    try:
        qd_elem = gen_quantdiff_elem(merged_ring)
        fastring.append(qd_elem)
    except ValueError:      # No synchrotron radiation => no diffusion element
        pass
    fastring = Lattice(fastring, **vars(ring))
    return fastring


def fast_ring(ring: Lattice, split_inds=[]) -> Tuple[Lattice, Lattice]:
    """Generates a "fast ring"

    A fast ring consisting in:

    * a RF cavity per distinct frequency,
    * a 6x6 linear transfer map,
    * a detuning and chromaticity element,
    * a quantum diffusion element (for radiation ring).

    2 new lattices are returned, one with radiation and one without
    It is possible to split the original ring in multiple "fastrings"
    using the ``split_inds`` argument

    Parameters:
        ring:       Lattice description
        split_inds: List of indexes where to split the ring

    Returns:
        fring (Lattice):    Fast ring without radiation
        fringrad (Lattice): Fast ring with radiation
    """
    ringi = ring.deepcopy()
    fastringnorad = _fring(ringi.radiation_off(copy=True),
                           split_inds=split_inds)
    detuning_elem = copy.deepcopy(get_elements(fastringnorad,
                                               'NonLinear')[0])
    fastringrad = _fring(ringi.radiation_on(copy=True),
                         split_inds=split_inds,
                         detuning_elem=detuning_elem)
    return fastringnorad, fastringrad
